#
# Copyright (C) 2021 The Delta Lake Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
try:
    from yarl import URL
    from yarl._quoting import _Quoter

    # Patch yarl.URL to not replace '%3D' with '=' which would break GCS pre-signed urls
    URL._PATH_REQUOTER = _Quoter(safe="@:", protected="/+=")  # type: ignore
except:
    pass
